import torch
import opt_einsum


def inner_product_vec(Z1, Z2=None):  # (M, D), (N, D)
    """
    Takes the pairwise inner products of all rows of Z1 and Z2.

    Args:
        Z1: first data matrix (M, D)
        Z2: second data matrix (N, D)

    Returns:
        A matrix containing the pairwise inner products (M, N)
    """
    if Z2 is None:
        Z2 = Z1
    return Z1.conj() @ Z2.T  # (M, N)


def inner_product_shift_invariant(Z1, Z2=None):  # (M, C, T), (N, C, T)
    """
    Takes the pairwise inner products of all rows of Z1 and Z2.

    Args:
        Z1: first data matrix (M, C, T)
        Z2: second data matrix (N, C, T)

    Returns:
        A matrix containing the pairwise inner products (M, N)
    """
    if Z2 is None:
        Z2 = Z1
    return opt_einsum.contract("mct, nct -> mn", Z1.conj(), Z2)  # (M, N)


def inner_product_translation_invariant(Z1, Z2=None):  # (M, C, H, W), (N, C, H, W)
    """
    Takes the pairwise inner products of all rows of Z1 and Z2.

    Args:
        Z1: first data matrix (M, C, H, W)
        Z2: second data matrix (N, C, H, W)

    Returns:
        A matrix containing the pairwise inner products (M, N)
    """
    if Z2 is None:
        Z2 = Z1
    return opt_einsum.contract("mchw, nchw -> mn", Z1.conj(), Z2)  # (M, N)


def gram_vec(Z):  # (N, D)
    """
    Computes the empirical second moment of a data matrix Z.
    Args:
        Z: a data matrix (N, D)

    Returns:
        The covariance matrix of Z data (D, D)
    """
    return Z.T @ Z.conj()  # (N, D) x (N, D) -> (D, D)

def gram_per_class_vec(Z, Pi):  # (N, D), (N, K)
    """
    Computes the empirical second moment of each class of a data matrix Z.
    Args:
        Z: a data matrix (N, D)
        Pi: a class information matrix (N, K)

    Returns:
        The class covariance tensor of Z data (K, D, D)
    """
    return opt_einsum.contract("ni, nj, nk -> kij", Z, Z.conj(), Pi.to(torch.float))  # (N, D) x (N, D) x (K, N) -> (K, D, D)

def gram_shift_invariant(Z):  # (N, C, T)
    """
    Computes the empirical second moment of a multi-channel 1D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last dimension.
    Args:
        Z: a data matrix (N, C, T)

    Returns:
        The covariance matrix of Z data (T, C, C)
    """
    return opt_einsum.contract("nct, ndt -> tcd", Z, Z.conj())  # (N, C, T) x (N, C, T) ->  (T, C, C)

def gram_per_class_shift_invariant(Z, Pi):  # (K, T, C, C),  (N, K)
    """
    Computes the empirical second moment of each class of a multi-channel 1D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last dimension.
    Args:
        Z: a data matrix (N, C, T)
        Pi: a class information matrix (N, K)

    Returns:
        The class covariance tensor of Z data (K, T, C, C)
    """
    return opt_einsum.contract("nct, ndt, nk -> ktcd", Z, Z.conj(), Pi.to(torch.float))  # (N, C, T) x (N, C, T) x (K, N) -> (K, T, C, C)

def gram_translation_invariant(Z):  # (N, C, H, W)
    """
    Computes the empirical second moment of a multi-channel 2D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last two dimensions.
    Args:
        Z: a data matrix (N, C, H, W)

    Returns:
        The covariance matrix of Z data (H, W, C, C)
    """
    return opt_einsum.contract("nchw, ndhw -> hwcd", Z, Z.conj())  # (N, C, H, W) x (N, C, H, W) -> (H, W, C, C)

def gram_per_class_translation_invariant(Z, Pi):  # (N, C, H, W), (N, K)
    """
    Computes the empirical second moment of each class of a multi-channel 2D data matrix Z.
    Should be used in contexts where Z matrix is already FFTed in last two dimensions.
    Args:
        Z: a data matrix (N, C, H, W)
        Pi: a class information matrix (N, K)

    Returns:
        The class covariance tensor of Z data (K, H, W, C, C)
    """
    return opt_einsum.contract("nchw, ndhw, nk -> khwcd", Z, Z.conj(), Pi.to(torch.float))  # (N, C, H, W) x (N, C, H, W) x (N, K) -> (K, H, W, C, C)

__all__ = [
    "inner_product_vec", "inner_product_shift_invariant", "inner_product_translation_invariant",
    "gram_vec", "gram_per_class_vec",
    "gram_shift_invariant", "gram_per_class_shift_invariant",
    "gram_translation_invariant", "gram_per_class_translation_invariant"
]



import mcr2
import torch


class SupervisedLDRLoss(torch.nn.Module):
    def __init__(self, eps_sq):
        super(SupervisedLDRLoss, self).__init__()
        self.cr = mcr2.coding_rate.SupervisedVectorCodingRate(eps_sq)


class SupervisedBinaryLDRLoss(SupervisedLDRLoss):
    def forward(self, Z, Z_hat):
        return self.cr.DeltaR_distance(Z, Z_hat)


class SupervisedMulticlassLDRLoss(SupervisedLDRLoss):
    def forward(self, Z, Z_hat, Pi):
        N = Z.shape[0]
        K = Pi.shape[1]
        RZ = self.cr.R(Z)
        RZ_hat = self.cr.R(Z_hat)
        RZ_Z_hat_concat = self.cr.R(torch.cat(tensors=(Z, Z_hat), dim=0))
        Z_per_class = [Z[Pi[:, i] == 1] for i in range(K)]
        Z_hat_per_class = [Z_hat[Pi[:, i] == 1] for i in range(K)]
        return RZ + RZ_hat + RZ_Z_hat_concat - sum(
            ((Z_per_class[i].shape[0] / N) + (1 / 2)) * self.cr.DeltaR_distance(Z_per_class[i], Z_hat_per_class[i]) for
            i in range(K)
        )

    def forward_unoptimized(self, Z, Z_hat, Pi):
        Z_per_class = [Z[Pi[:, i] == 1] for i in range(Pi.shape[1])]
        Z_hat_per_class = [Z_hat[Pi[:, i] == 1] for i in range(Pi.shape[1])]
        return self.cr(Z) + self.cr(Z_hat) + \
               sum(self.cr.DeltaR_distance(Z_per_class[i], Z_hat_per_class[i]) for i in range(Pi.shape[1]))


__all__ = ["SupervisedBinaryLDRLoss", "SupervisedMulticlassLDRLoss"]



import mcr2.functional as F
import torch


class SupervisedCodingRate:
    def __init__(self, eps_sq):
        self.eps_sq = eps_sq

    def R(self, Z):
        raise NotImplementedError("Called R in basic coding rate class")

    def Rc(self, Z, Pi):
        raise NotImplementedError("Called Rc in basic coding rate class")

    def DeltaR(self, Z, Pi):
        return self.R(Z) - self.Rc(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        N1 = Z1.shape[0]
        N2 = Z2.shape[0]
        N = N1 + N2
        return self.R(torch.cat(tensors=(Z1, Z2), dim=0)) - (N1 / N) * self.R(Z1) - (N2 / N) * self.R(Z2)


class SupervisedVectorCodingRate(SupervisedCodingRate):
    def R(self, Z):
        """
        Computes the coding rate of Z.

        Args:
            Z: data matrix, (N x D)

        Returns:
            The coding rate of Z.
        """
        N, D = Z.shape  # (N, D)
        ZTZ = F.gram_vec(Z)  # (D, D)
        alpha = D / (N * self.eps_sq)
        I = torch.eye(D)  # (D, D)
        Sigma_hat = I + alpha * ZTZ  # (D, D)
        return F.logdet(Sigma_hat) / 2.0  # ()

    def Rc(self, Z, Pi):
        """
        Computes the segmented coding rate of Z with respect to a class information matrix Pi.

        Args:
            Z: data matrix, (N x D)
            Pi: class information matrix, (K x N)

        Returns:
            The segmented coding rate of Z with respect to Pi.
        """
        N, D = Z.shape  # (N, D)
        N, K = Pi.shape  # (N, K)
        ZTZ_per_class = F.gram_per_class_vec(Z, Pi)  # (K, D, D)
        N_per_class = torch.sum(Pi, axis=0)  # (K, )
        gamma_per_class = N_per_class / N  # (K, )
        alpha_per_class = torch.where(  # stops divide by 0 errors
            N_per_class > 0.0,
            D / (self.eps_sq * N_per_class),  # (K, )
            torch.tensor(0.0)  # ()
        )  # (K, )
        I = torch.eye(D).view(1, D, D)  # (D, D)
        Sigma_hat_per_class = I + alpha_per_class.view(K, 1, 1) * ZTZ_per_class  # (K, D, D)
        logdets_per_class = F.logdet(Sigma_hat_per_class)  # (K, )
        return torch.sum(gamma_per_class * logdets_per_class) / 2.0  # ()

    def DeltaR(self, Z, Pi):
        """
        Computes the coding rate reduction of Z with respect to a class information matrix Pi.

        Args:
            Z: data matrix, (N x D)
            Pi: class information matrix, (K x N)

        Returns:
            The coding rate reduction of Z with respect to Pi.
        """
        return super().DeltaR(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        """
        Computes the DeltaR distance between the two data point sets Z1 and Z2.

        Args:
            Z1: data matrix, (N1 x D)
            Z2: data matrix, (N2 x D)

        Returns:
            The DeltaR distance between Z1 and Z2.
        """
        return super().DeltaR_distance(Z1, Z2)


class SupervisedShiftInvariantCodingRate(SupervisedCodingRate):
    def R(self, Z):
        """
        Computes the coding rate of Z.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N, C, T)

        Returns:
            The coding rate of Z.
        """
        N, C, T = Z.shape  # (N, C, T)
        ZTZ = F.gram_shift_invariant(Z)  # (T, C, C)
        alpha = C / (N * self.eps_sq)
        I = torch.eye(C).view(1, C, C)  # (1, C, C)
        Sigma_hat = I + alpha * ZTZ  # (T, C, C)
        return torch.sum(F.logdet(Sigma_hat)) / 2.0  # ()

    def Rc(self, Z, Pi):
        """
        Computes the segmented coding rate of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N, C, T)
            Pi: class information matrix, (N, K)

        Returns:
            The segmented coding rate of Z with respect to Pi.
        """
        N, C, T = Z.shape  # (N, C, T)
        N, K = Pi.shape  # (N, K)
        ZTZ_per_class = F.gram_per_class_shift_invariant(Z, Pi)  # (K, T, C, C)
        N_per_class = torch.sum(Pi, axis=0)  # (K, )
        gamma_per_class = N_per_class / N  # (K, )
        alpha_per_class = torch.where(  # stops divide by 0 errors
            N_per_class > 0,
            C / (self.eps_sq * N_per_class),  # (K, )
            torch.tensor(0.0)  # ()
        )  # (K, )
        I = torch.eye(C).view(1, 1, C, C)  # (1, 1, C, C)
        Sigma_hat_per_class = I + alpha_per_class.view(K, 1, 1, 1) * ZTZ_per_class  # (K, T, C, C)
        logdets_per_class = torch.sum(
            F.logdet(Sigma_hat_per_class),  # (K, T)
            dim=1
        )  # (K, )
        return torch.sum(gamma_per_class * logdets_per_class) / 2.0  # ()

    def DeltaR(self, Z, Pi):
        """
        Computes the coding rate reduction of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z: data matrix, (N x C x T)
            Pi: class information matrix, (K x N)

        Returns:
            The coding rate reduction of Z with respect to Pi.
        """
        return super().DeltaR(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        """
        Computes the DeltaR distance between the two data point sets Z1 and Z2.
        NOTE: Only an accurate measure of the distance if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft(Z).

        Args:
            Z1: data matrix, (N1 x C x T)
            Z2: data matrix, (N2 x C x T)

        Returns:
            The DeltaR distance between Z1 and Z2.
        """
        return super().DeltaR_distance(Z1, Z2)


class SupervisedTranslationInvariantCodingRate(SupervisedCodingRate):
    def R(self, Z):
        """
        Computes the coding rate of Z.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)

        Returns:
            The coding rate of Z.
        """
        N, C, H, W = Z.shape  # (N, C, H, W)
        ZTZ = F.gram_translation_invariant(Z)  # (H, W, C, C)
        alpha = C / (N * self.eps_sq)
        I = torch.eye(C).view(1, 1, C, C)  # (1, 1, C, C)
        Sigma_hat = I + alpha * ZTZ  # (H, W, C, C)
        return torch.sum(F.logdet(Sigma_hat)) / 2.0  # ()

    def Rc(self, Z, Pi):
        """
        Computes the segmented coding rate of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)
            Pi: class information matrix, (N, K)

        Returns:
            The segmented coding rate of Z with respect to Pi.
        """
        N, C, H, W = Z.shape  # (N, C, H, W)
        N, K = Pi.shape  # (N, K)
        ZTZ_per_class = F.gram_per_class_translation_invariant(Z, Pi)  # (K, H, W, C, C)
        N_per_class = torch.sum(Pi, axis=0)  # (K, )
        gamma_per_class = N_per_class / N  # (K, )
        alpha_per_class = torch.where(  # stops divide by 0 errors
            N_per_class > 0,
            C / (self.eps_sq * N_per_class),  # (K, )
            torch.tensor(0.0)  # ()
        )  # (K, )
        I = torch.eye(C).view(1, 1, 1, C, C)  # (1, 1, 1, C, C)
        Sigma_hat_per_class = I + alpha_per_class.view(K, 1, 1, 1, 1) * ZTZ_per_class  # (K, H, W, C, C)
        logdets_per_class = torch.sum(
            F.logdet(Sigma_hat_per_class),  # (K, H, W)
            dim=(1, 2)
        )  # (K, )
        return torch.sum(gamma_per_class * logdets_per_class) / 2.0  # ()

    def DeltaR(self, Z, Pi):
        """
        Computes the coding rate reduction of Z with respect to a class information matrix Pi.
        NOTE: Only an accurate measure of the coding rate if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z: data matrix, (N, C, H, W)
            Pi: class information matrix, (N, K)

        Returns:
            The coding rate reduction of Z with respect to Pi.
        """
        return super().DeltaR(Z, Pi)

    def DeltaR_distance(self, Z1, Z2):
        """
        Computes the DeltaR distance between the two data point sets Z1 and Z2.
        NOTE: Only an accurate measure of the distance if the data is FFTed beforehand.
        You can do this with mcr2.functional.fft2(Z).

        Args:
            Z1: data matrix, (N1 x C x H x W)
            Z2: data matrix, (N2 x C x H x W)

        Returns:
            The DeltaR distance between Z1 and Z2.
        """
        return super().DeltaR_distance(Z1, Z2)


__all__ = [
    "SupervisedVectorCodingRate",
    "SupervisedShiftInvariantCodingRate",
    "SupervisedTranslationInvariantCodingRate"
]



import torch
import pytorch_lightning


class ReduLayer(pytorch_lightning.LightningModule):
    def __init__(self, eta, eps_sq, lmbda, num_classes, differentiable):
        super(ReduLayer, self).__init__()
        self.E = None
        self.Cs = None
        self.gam = None

        self.eta = eta
        self.eps_sq = eps_sq
        self.lmbda = lmbda

        self.K = num_classes

        self.diff = differentiable
        self.diff_ctx = lambda: torch.no_grad() if not self.diff else torch.enable_grad()

    def forward(self, V):
        raise NotImplementedError("Attempted to call forward with abstract ReduLayer")

    def zero(self):
        self.set_params(torch.zeros_like(self.E), torch.zeros_like(self.Cs))

    def init(self, V, Pi):
        gam = self.compute_gam(Pi)
        E = self.compute_E(V)
        Cs = self.compute_Cs(V, Pi)
        self.set_params(E, Cs, gam)

    def update(self, V, Pi, tau):
        E_ref, Cs_ref = self.E, self.Cs
        E_new = self.compute_E(V)
        Cs_new = self.compute_Cs(V, Pi)
        E_update = E_ref + tau * (E_new - E_ref)
        Cs_update = Cs_ref + tau * (Cs_new - Cs_ref)
        self.set_params(E_update, Cs_update)

    def set_params(self, E, Cs, gam=None):
        self.E = torch.nn.Parameter(E, requires_grad=self.E.requires_grad)
        self.Cs = torch.nn.Parameter(Cs, requires_grad=self.Cs.requires_grad)
        if gam is not None:
            self.gam = torch.nn.Parameter(gam, requires_grad=False)

__all__ = ["ReduLayer"]



import torch

def logdet(H):
    """
    Takes the log determinant of a (possible batch of) Hermitian matrix.
    Args:
        H: A Hermitian matrix batch (*, D, D)

    Returns:
        The log determinant of H (*, )
    """
    L = torch.linalg.cholesky(H)  # (*, D, D)
    diag_L = torch.diagonal(L, dim1=-2, dim2=-1)  # (*, D)
    log_diag_L = torch.log(diag_L)  # (*, D)
    tr_log_L = torch.sum(log_diag_L, dim=-1)  # (*, )
    return 2 * tr_log_L  # (*, )

__all__ = ["logdet"]


import torch
import mcr2.functional as F

class ReduBlock(torch.nn.Module):
    # ReduNet Architecture. This class is used to stack a series of ReduLayers.

    def __init__(self, *layers):
        super(ReduBlock, self).__init__()
        self.layers = torch.nn.ModuleList(layers)
        self.losses = []

    def init(self, X, Pi):
        # Initialize the network. Using inputs and labels, it constructs
        # the parameters E and Cs throughout each ReduLayer.
        return self.forward(X, Pi,
                            init=True,
                            loss=True)

    def update(self, X, Pi, tau=0.1):
        # Update the network parameters E and Cs by
        # performing a moving average.
        return self.forward(X, Pi, tau=tau,
                            update=True,
                            loss=True)

    def zero(self):
        # Set every network parameters E and Cs to a zero matrix.
        for layer in self.layers:
            layer.zero()

    def forward(self, X, Pi=None, tau=0.1,
                init=False,
                update=False,
                loss=False):
        for idx, layer in enumerate(self.layers):
            # preprocess for redunet layers
            if idx == 0:
                X = layer.preprocess(X)

            # If init is set to True, then initialize
            # layer using inputs and labels
            if init:
                layer.init(X, Pi)

            # If update is set to True, then update
            # layer using inputs and labels
            if update:
                layer.update(X, Pi, tau)

            # Perform a forward pass
            X = layer.forward(X)

            # compute MCR quantities for redunet layer
            if loss:
                mcr_losses = layer.compute_mcr2(X, Pi)
                self._log_losses(idx, *mcr_losses)

            # postprocess for redunet layers
            if idx == len(self.layers) - 1:
                X = layer.postprocess(X)
        return X

    def _log_losses(self, layer_idx, DeltaR, R, Rc):
        self.losses.append(
            {"layer_idx": layer_idx, "DeltaR": DeltaR, "R": R, "Rc": Rc}
        )

    def get_losses_per_layer(self):
        return self.losses

__all__ = ["ReduBlock"]



import torch
import mcr2
import mcr2.functional as F
import opt_einsum

from .abstract import ReduLayer


class TranslationInvariantReduLayer(ReduLayer):
    def __init__(self, eta, eps_sq, lmbda, num_classes, dimensions, differentiable):
        super(TranslationInvariantReduLayer, self).__init__(eta, eps_sq, lmbda, num_classes, differentiable)

        self.C, self.H, self.W = dimensions

        self.gam = torch.nn.Parameter(
            torch.ones(self.K) / self.K,  # (K, )
            requires_grad=False
        )
        self.E = torch.nn.parameter.Parameter(
            torch.empty(size=(self.H, self.W, self.C, self.C)),  # (D, D)
            requires_grad=self.diff
        )
        self.Cs = torch.nn.parameter.Parameter(
            torch.empty(size=(self.K, self.H, self.W, self.C, self.C)),  # (K, D, D)
            requires_grad=self.diff
        )

        self.cr = mcr2.coding_rate.SupervisedTranslationInvariantCodingRate(eps_sq)

    def forward(self, V):  # (N, C, H, W)
        with self.diff_ctx():
            expd = opt_einsum.contract("nchw, hwcd -> ndhw", V, self.E.conj())  # (N, C, H, W) x (H, W, C, C) -> (N, C, H, W)
            comp = opt_einsum.contract("nchw, khwcd -> kndhw", V, self.Cs.conj())  # (N, C, H, W) x (K, H, W, C, C) -> (K, N, C, H, W)
            clus = self.nonlinear(comp)  # (N, C, H, W)
            V = V + self.eta * (expd - clus)  # (N, C, H, W)
            return F.normalize_2d(V)  # (N, C, H, W)

    def nonlinear(self, Cz):  # (K, N, C, H, W)
        K, N, C, H, W = Cz.shape  # (K, N, C, H, W)
        norm = torch.linalg.norm(Cz.flatten(start_dim=2), axis=2)  # (K, N)
        pred = torch.nn.functional.softmax(-self.lmbda * norm, dim=0).view(K, N, 1, 1, 1)  # (K, N, 1, 1, 1)
        gam = self.gam.view(K, 1, 1, 1, 1)  # (K, 1, 1, 1, 1)
        out = torch.sum(gam * Cz * pred, axis=0)  # sum((K, 1, 1, 1, 1) x (K, N, C, H, W) x (K, N, 1, 1, 1) -> (K, N, C, H, W)) -> (N, C, H, W)
        return out  # (N, C, H, W)

    def compute_gam(self, Pi):  # (K, N)
        with torch.no_grad():
            K, N = Pi.shape  # (K, N)
            N_per_class = torch.sum(Pi, axis=0)  # (K, )
            gamma_per_class = N_per_class / N  # (K, )
            return gamma_per_class  # (K, )

    def compute_E(self, V):  # (N, C, H, W)
        with self.diff_ctx():
            N, C, H, W = V.shape  # (N, C, H, W)
            VTV = F.gram_translation_invariant(V)  # (H, W, C, C)
            alpha = C / (N * self.eps_sq)
            I = torch.eye(C).view(1, 1, C, C)  # (1, 1, C, C)
            Sigma_hat = I + alpha * VTV  # (H, W, C, C)
            return alpha * torch.linalg.inv(Sigma_hat)  # (H, W, C, C)

    def compute_Cs(self, V, Pi):  # (N, C, H, W), (K, N)
        with self.diff_ctx():
            N, C, H, W = V.shape  # (N, C, H, W)
            N, K = Pi.shape  # (K, N)
            VTV_per_class = F.gram_per_class_translation_invariant(V, Pi)  # (K, H, W, C, C)
            N_per_class = torch.sum(Pi, axis=0)  # (K, )
            gamma_per_class = N_per_class / N  # (K, )
            alpha_per_class = torch.where(  # stops divide by 0 errors
                N_per_class > 0.0,
                C / (self.eps_sq * N_per_class),  # (K, )
                torch.tensor(0.0)  # ()
            )  # (K, )
            alpha_per_class = alpha_per_class.view(K, 1, 1, 1, 1)  # (K, 1, 1, 1, 1)
            I = torch.eye(C).view(1, 1, 1, C, C)  # (1, 1, 1, C, C)
            Sigma_hat_per_class = I + alpha_per_class * VTV_per_class  # (K, H, W, C, C)
            return alpha_per_class * torch.linalg.inv(Sigma_hat_per_class)  # (K, H, W, C, C)

    def preprocess(self, V):
        with self.diff_ctx():
            return F.normalize_2d(F.fft2(V))

    def postprocess(self, V):
        with self.diff_ctx():
            return F.normalize_2d(F.ifft2(V).real.to(torch.float))

    def compute_mcr2(self, Z, Pi):
        with torch.no_grad():
            return (self.cr.DeltaR(Z, Pi), self.cr.R(Z), self.cr.Rc(Z, Pi))

__all__ = ["TranslationInvariantReduLayer"]
